/*
 * Copyright (C) 2009 SHARP Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.co.sharp.android.SampleCamera;

import java.util.Arrays;

/**
 * MPフォーマットのデータ生成機能を提供するクラスである。
 * 本クラスを作成するにあたり以下の規格書を参考にしている。
 *   ・デジタルスチルカメラ用画像ファイルフォーマット規格Exif 2.3
 *       http://cipa.jp/hyoujunka/kikaku/pdf/DC-008-2010_J.pdf
 *   ・マルチピクチャフォーマット
 *       http://cipa.jp/hyoujunka/kikaku/pdf/DC-007_J.pdf
 */
public class MpfUtility
{
    /* APPマーカ種別 */
    private static final int APP1_EXIF      = 0;
    private static final int APP2_MPF       = 1;

    /* APPマーカセグメント */
    private static final byte[] MARKER_SOI   = {(byte)0xFF, (byte)0xD8};
    private static final byte[] MARKER_APP1  = {(byte)0xFF, (byte)0xE1};
    private static final byte[] MARKER_APP2  = {(byte)0xFF, (byte)0xE2};

    /* フォーマット識別値 */
    private static final byte[] FORMAT_MPF   = {'M', 'P', 'F', 0};
    private static final byte[] FORMAT_EXIF  = {'E', 'x', 'i', 'f', 0};

    /*--- MPフォーマット情報 各種テーブル ---*/
    
    /* MPフォーマット付属情報（先頭のみ） */
    private final byte[] infoApp2Header = {
                /* APP2マーカ */
                (byte)0xFF, (byte)0xE2,
                /* APP2マーカセグメント長 */
    /*  +2 */   (byte)0x00, (byte)0x00,                             /* ★編集必要 */
                /* MPフォーマット識別コード */
                (byte)0x4D, (byte)0x50, (byte)0x46, (byte)0x00      /* MPF */
    };
    private static final int OFS_APPMARKER_LENGTH = 2;
    private static final int OFS_APPMARKER_CODE = 4;
    
    /* MPヘッダ */
    private final byte[] infoMpHeader = {
                /* MPエンディアン情報 */
                (byte)0x49, (byte)0x49, (byte)0x2A, (byte)0x00,     /* little endian */
                /* 先頭IFDへのオフセット */
                (byte)0x08, (byte)0x00, (byte)0x00, (byte)0x00
    };

    /* MPインデックスIFD */
    private final byte[] infoMpIndexIfd = {
                /* カウント */
                (byte)0x03, (byte)0x00,                             /* 3 */
                /* MPフォーマットバージョンタグ */
                (byte)0x00, (byte)0xB0,
                (byte)0x07, (byte)0x00,
                (byte)0x04, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x30, (byte)0x31, (byte)0x30, (byte)0x30,
                /* 記録画像数タグ */
                (byte)0x01, (byte)0xB0,
                (byte)0x04, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +22 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* MPエントリタグ */
                (byte)0x02, (byte)0xB0,
                (byte)0x07, (byte)0x00,
    /* +30 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                (byte)0x32, (byte)0x00, (byte)0x00, (byte)0x00,
                /* Offset of Next IFD */
    /* +38 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00      /* ★編集必要 */
    };
    private static final int OFS_MPIDXIFD_IMAGES_COUNT = 22;
    private static final int OFS_MPIDXIFD_ENTRY_COUNT  = 30;
    private static final int OFS_MPIDXIFD_NEXT_OFFSET   = 38;
    private static final int LEN_MPENTRY  = 16;   /* MPエントリサイズ */

    /* MPエントリ */
    private final byte[] infoMpEntry = {
                /* 個別画像種別管理情報 */
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 個別画像サイズ */
    /* +4 */    (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 個別画像データオフセット */
    /* +8 */    (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 従属画像1 エントリ番号 */
                (byte)0x00, (byte)0x00,
                /* 従属画像2 エントリ番号 */
                (byte)0x00, (byte)0x00
    };
    private static final int OFS_MPENTRY_KIND   = 0;
    private static final int OFS_MPENTRY_LENGTH = 4;
    private static final int OFS_MPENTRY_OFFSET = 8;
    /* 個別画像種別 */
    private static final int MPENTRY_KIND_MASTER = 0x20020002;  /* 代表画像 */
    private static final int MPENTRY_KIND_SLAVE  = 0x00020002;  /* 従属画像 */

    /* MP個別情報IFD（先頭画像用） */
    private final byte[] infoMpIndividualIfd0th = {
                /* カウント */
                (byte)0x04, (byte)0x00,
                /* 個別画像番号タグ */
                (byte)0x01, (byte)0xB1,
                (byte)0x04, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +10 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 基準視点番号タグ */
                (byte)0x04, (byte)0xB2,
                (byte)0x04, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                /* 輻輳角タグ */
                (byte)0x05, (byte)0xB2,
                (byte)0x0A, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +34 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 基線長タグ */
                (byte)0x06, (byte)0xB2,
                (byte)0x05, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +46 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* Offset of Next IFD */
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
                /* value - 輻輳角：0度 */
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                /* value - 基線長：不定 */
                (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
                (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF
    };
    private static final int OFS_MPINDIFD0_INDNUM_OFFSET     = 10;
    private static final int OFS_MPINDIFD0_CONVANGLE_OFFSET  = 34;
    private static final int OFS_MPINDIFD0_BASELENGTH_OFFSET = 46;

    /* MP個別情報IFD（先頭以外の画像用） */
    private final byte[] infoMpIndividualIfd = {
                /* カウント */
                (byte)0x05, (byte)0x00,
                /* MPFフォーマットバージョンタグ */
                (byte)0x00, (byte)0xB0,
                (byte)0x07, (byte)0x00,
                (byte)0x04, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x30, (byte)0x31, (byte)0x30, (byte)0x30,
                /* 個別画像番号タグ */
                (byte)0x01, (byte)0xB1,
                (byte)0x04, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +22 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 基準視点番号タグ */
                (byte)0x04, (byte)0xB2,
                (byte)0x04, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                /* 輻輳角タグ */
                (byte)0x05, (byte)0xB2,
                (byte)0x0A, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +46 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* 基線長タグ */
                (byte)0x06, (byte)0xB2,
                (byte)0x05, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
    /* +58 */   (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,     /* ★編集必要 */
                /* Offset of Next IFD */
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
                /* value - 輻輳角：0度 */
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
                (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00,
                /* value - 基線長：不定 */
                (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
                (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF
    };
    private static final int OFS_MPINDIFD_INDNUM_OFFSET     = 22;
    private static final int OFS_MPINDIFD_CONVANGLE_OFFSET  = 46;
    private static final int OFS_MPINDIFD_BASELENGTH_OFFSET = 58;

    /* SRATIONAL型valueサイズ */
    private static final int LEN_VAL_SRATIONAL = 8; 

    /**
     * MPOデータを生成する。
     * 複数のJPEGデータからMPフォーマット形式のMPOデータを生成する。
     * 立体視マルチビュー画像のみ生成可。
     * @param jpeg JPEGデータ
     * @return MPOデータ
     */
    public byte[] createMpfData(byte[][] jpeg)
    {
        /* 引数チェック */
        int num = jpeg.length;
        if((jpeg == null) || (num < 1)) {
            return null;
        }

        /* 出力バッファを確保 */
        int outLength = jpeg[0].length                      /* APP1/Data */
                        + infoApp2Header.length             /* APP2 - header */
                        + infoMpHeader.length               /* APP2 - MP header */
                        + infoMpIndexIfd.length             /* APP2 - MP indexIFD */
                        + infoMpEntry.length * jpeg.length  /* APP2 - MP indexIFD value(0th) */
                        + infoMpIndividualIfd0th.length;    /* APP2 - MP individualIFD(0th) */
        for(int i = 1; i < num; i++) {
            outLength += jpeg[i].length                     /* APP1/Data */
                            + infoApp2Header.length         /* APP2 - header */
                            + infoMpHeader.length           /* APP2 - MP header */
                            + infoMpIndividualIfd.length;   /* APP2 - MP individualIFD */
        }
        byte[] outBuff = new byte[outLength];

        /* 書き込み済みデータサイズ */
        int writeSize = 0;

        /* JPEG画像ごとにMPO形式に変換 */
        for(int index = 0; index < num; index++) {
            byte[] inBuff = jpeg[index];
            int length = inBuff.length;

            /* APP1マーカ(Exif)を検索 */
            int[] result = searchAppMarker(inBuff, 0, APP1_EXIF);
            int offsetAPP1 = result[0];
            int lengthAPP1 = result[1];

            /* 先頭画像を変換 */
            if(index == 0) {
                /* APP1(Exif)をバッファに書き込む */
                writeSize += copyBuffer(inBuff, 0, outBuff, writeSize,
                                        offsetAPP1 + lengthAPP1, "APP1-Exif0");
                /* APP2(MPF)をバッファに書き込む */
                writeSize = addApp2Marker0th(outBuff, writeSize, num, length);
            }
            /* 先頭以外の画像を変換 */
            else {
                int offsetTop = writeSize;

                /* 先頭画像のAPP2マーカ(MPF)を検索 */
                result = searchAppMarker(outBuff, 0, APP2_MPF);
                int offsetAPP2 = result[0];

                /* APP1(Exif)をバッファに書き込む */
                writeSize += copyBuffer(inBuff, 0, outBuff, writeSize,
                                        offsetAPP1 + lengthAPP1, "APP1-Exif" + index);
                /* APP2(MPF)をバッファに書き込む */
                writeSize = addApp2Marker(outBuff, writeSize, index, length,
                                                offsetAPP2, offsetTop);
            }

            /* 個別画像データをバッファに書き込む */
            writeSize += copyBuffer(inBuff, offsetAPP1 + lengthAPP1, outBuff, writeSize,
                                    length - (offsetAPP1 + lengthAPP1), "jpeg" + index);
        }
        
        return outBuff;
    }

    /**
     * APPマーカを検索する。
     * APP1(Exif)が検索対象であった場合は、その直前にあるSOIマーカを含めて検索結果とする。
     * APP1(Exif)/APP2(MPF)以外のAPPマーカには非対応。
     * @param buff 検索対象データ
     * @param offset 検索開始位置
     * @param target 検索するAPPマーカ種別
     * @return [0]…APPマーカ先頭アドレスへのオフセット(APP1の場合はSOIのオフセット）
     *         [1]…APPマーカのサイズ（APP1の場合はSOIを含めたサイズ）
     */
    private int[] searchAppMarker(byte[] buff, int offset, int target)
    {
        /* SOIマーカのチェック */
        if(!containsMarker(buff, offset, MARKER_SOI, "SOI")) {
            throw new RuntimeException("Invalid data : not found SOI");
        }
        /* オフセットを進める */
        offset += MARKER_SOI.length;

        /* APP1マーカのチェック */
        if(!containsMarker(buff, offset, MARKER_APP1, "APP1")) {
            throw new RuntimeException("Invalid data : not found APP1 marker");
        }
        /* APP1フォーマットコードのチェック */
        if(!containsMarker(buff, offset + OFS_APPMARKER_CODE, FORMAT_EXIF, "APP1 format")) {
            throw new RuntimeException("Invalid data : not found APP1 format code");
        }

        /* APP1マーカサイズを取得 */
        int pos = offset + OFS_APPMARKER_LENGTH;
        int size = MARKER_APP1.length
                    + (((int)buff[pos  ] << 8) & 0xFF00)
                    + ( (int)buff[pos+1]       & 0x00FF);

        /* 検索対象がAPP1ならば結果を返却して終了 */
        if(target == APP1_EXIF) {
            /* SOIを反映したオフセット/サイズを返却 */
            int ret[] = {offset - MARKER_SOI.length, size + MARKER_SOI.length};
            return ret;
        }
        
        /* 次マーカへオフセットを進める */
        offset += size;

        /* APP2マーカのチェック */
        if(!containsMarker(buff, offset, MARKER_APP2, "APP2")) {
            throw new RuntimeException("Invalid data : not found APP2 marker");
        }
        /* APP2フォーマットコードのチェック */
        if(!containsMarker(buff, offset + OFS_APPMARKER_CODE, FORMAT_MPF, "APP2 format")) {
            throw new RuntimeException("Invalid data : not found APP2 format code");
        }

        /* APP2マーカサイズを取得 */
        pos = offset + OFS_APPMARKER_LENGTH;
        size = MARKER_APP2.length
                + (((int)buff[pos  ] << 8) & 0xFF00)
                + ( (int)buff[pos+1]       & 0x00FF);

        /* 検索対象がAPP2ならば結果を返却して終了 */
        if(target == APP2_MPF) {
            int ret[] = {offset, size};
            return ret;
        }

        throw new RuntimeException("Invalid argument : kind(" + target + ")");
    }

    /**
     * 先頭画像のAPP2マーカ情報をバッファに追加する。
     * @param buff 出力バッファ
     * @param offset 出力開始位置
     * @param num 個別画像数
     * @param size 個別画像サイズ
     * @return
     */
    private int addApp2Marker0th(byte[] buff, int offset, int num, int size)
    {
        int offsetTop = offset;
        int value = 0;

        /*--- MPフォーマット付属情報（ヘッダ部） ---*//* 設定は後半で */

        /* バッファに書き込む */
        offset += copyBuffer(infoApp2Header, 0, buff, offset, "APP2 header");

        /*--- MPヘッダ ---*/

        /* バッファに書き込む */
        offset += copyBuffer(infoMpHeader, 0, buff, offset, "APP2 MPHeader");

        /*--- MPインデックスIFD ---*/

        /* 記録画像数 */
        setLongValue(infoMpIndexIfd, OFS_MPIDXIFD_IMAGES_COUNT, num);
        /* MPエントリ */
        setLongValue(infoMpIndexIfd, OFS_MPIDXIFD_ENTRY_COUNT, num * LEN_MPENTRY);
        /* Offset of Next IFD */
        value = infoMpHeader.length + infoMpIndexIfd.length + (num * LEN_MPENTRY);
        setLongValue(infoMpIndexIfd, OFS_MPIDXIFD_NEXT_OFFSET, value);
        /* バッファに書き込む */
        offset += copyBuffer(infoMpIndexIfd, 0, buff, offset, "APP2 MPIndexIFD");

        /*--- MPエントリ ---*//* 設定は後半で */

        /* バッファに書き込む */
        Arrays.fill(infoMpEntry, (byte)0x00);
        for(int index = 0; index < num; index++) {
            offset += copyBuffer(infoMpEntry, 0, buff, offset, "APP2 MPEntry-" + index);
        }

        /*--- MP個別情報IFD ---*/

        /* 個別画像番号タグ - value */
        setLongValue(infoMpIndividualIfd0th, OFS_MPINDIFD0_INDNUM_OFFSET, 1);
        /* 輻輳角タグ - オフセット */
        value = infoMpHeader.length + infoMpIndexIfd.length + (num * LEN_MPENTRY)
                + infoMpIndividualIfd0th.length - (2 * LEN_VAL_SRATIONAL);
        setLongValue(infoMpIndividualIfd0th, OFS_MPINDIFD0_CONVANGLE_OFFSET, value);
        /* 基線長タグ - オフセット */
        value += LEN_VAL_SRATIONAL;
        setLongValue(infoMpIndividualIfd0th, OFS_MPINDIFD0_BASELENGTH_OFFSET, value);
        /* バッファに書き込む */
        offset += copyBuffer(infoMpIndividualIfd0th, 0, buff, offset, "APP2 MPIndividualIFD-0");

        /* ここまででAPP2マーカ情報は一通り書き込み完了 */
        /* ここから一部設定値を書き換える */
        
        /*--- MPフォーマット付属情報（ヘッダ部） ---*/

        /* APP2マーカセグメント長(Big Endian) */
        value = offset - (offsetTop + OFS_APPMARKER_LENGTH);
        infoApp2Header[OFS_APPMARKER_LENGTH+1] = (byte)( ((short)value)       & 0xFF);
        infoApp2Header[OFS_APPMARKER_LENGTH  ] = (byte)((((short)value) >> 8) & 0xFF);
        /* バッファを書き換える。 */
        copyBuffer(infoApp2Header, 0, buff, offsetTop, "APP2 marker");

        /*--- MPエントリ（先頭画像） ---*/
        
        /* 個別画像種別管理情報 */
        setLongValue(infoMpEntry, OFS_MPENTRY_KIND, MPENTRY_KIND_MASTER);
        /* 個別画像サイズ */
        value = (offset - offsetTop) + size;
        setLongValue(infoMpEntry, OFS_MPENTRY_LENGTH, value);
        /* 個別画像データオフセット */
        setLongValue(infoMpEntry, OFS_MPENTRY_OFFSET, 0);
        /* バッファを書き換える。 */
        value = offsetTop + infoApp2Header.length + infoMpHeader.length
                + infoMpIndexIfd.length;
        copyBuffer(infoMpEntry, 0, buff, value, "APP2 MPEntry-0");

        return offset;
    }

    /**
     * 先頭以外の画像のAPP2マーカ情報をバッファに追加する。
     * @param buff 出力バッファ
     * @param offset 出力開始位置
     * @param index 個別画像インデックス
     * @param size 個別画像サイズ
     * @param offset0th 先頭画像のAPP2マーカへのオフセット
     * @param offsetMe 個別画像先頭(SOI)へのオフセット
     * @return
     */
    private int addApp2Marker(byte[] buff, int offset, int index, int size,
                                int offset0th, int offsetMe)
    {
        int offsetTop = offset;
        int value = 0;

        /*--- MPフォーマット付属情報（ヘッダ部） ---*//* 設定は後半で */

        /* バッファに書き込む */
        offset += copyBuffer(infoApp2Header, 0, buff, offset, "APP2 header-" + index);

        /*--- MPヘッダ ---*/

        /* バッファに書き込む */
        offset += copyBuffer(infoMpHeader, 0, buff, offset, "APP2 MPHeader-" + index);

        /*--- MP個別情報IFD ---*/

        /* 個別画像番号タグ - value */
        setLongValue(infoMpIndividualIfd, OFS_MPINDIFD_INDNUM_OFFSET, index + 1);
        /* 輻輳角タグ - オフセット */
        value = infoMpHeader.length + infoMpIndividualIfd.length - (2 * LEN_VAL_SRATIONAL);
        setLongValue(infoMpIndividualIfd, OFS_MPINDIFD_CONVANGLE_OFFSET, value);
        /* 基線長タグ - オフセット */
        value += LEN_VAL_SRATIONAL;
        setLongValue(infoMpIndividualIfd, OFS_MPINDIFD_BASELENGTH_OFFSET, value);
        /* バッファに書き込む */
        offset += copyBuffer(infoMpIndividualIfd, 0, buff, offset, "APP2 MPIndividualIFD-" + index);

        /* ここまででAPP2マーカ情報は一通り書き込み完了 */
        /* ここから一部設定値を書き換える */

        /*--- MPフォーマット付属情報（ヘッダ部） ---*/

        /* APP2マーカセグメント長(Big Endian) */
        value = offset - (offsetTop + OFS_APPMARKER_LENGTH);
        infoApp2Header[OFS_APPMARKER_LENGTH+1] = (byte)( ((short)value)       & 0xFF);
        infoApp2Header[OFS_APPMARKER_LENGTH  ] = (byte)((((short)value) >> 8) & 0xFF);
        /* バッファを書き換える。 */
        copyBuffer(infoApp2Header, 0, buff, offsetTop, "APP2 marker");

        /*--- MPエントリ（対象画像） ---*/
        
        /* 個別画像種別管理情報 */
        setLongValue(infoMpEntry, OFS_MPENTRY_KIND, MPENTRY_KIND_SLAVE);
        /* 個別画像サイズ */
        value = (offset - offsetTop) + size;
        setLongValue(infoMpEntry, OFS_MPENTRY_LENGTH, value);
        /* 個別画像データオフセット */
        value = offsetMe - (offset0th + infoApp2Header.length);
        setLongValue(infoMpEntry, OFS_MPENTRY_OFFSET, value);
        /* バッファを書き換える。 */
        value = offset0th + infoApp2Header.length + infoMpHeader.length
                + infoMpIndexIfd.length + (index * infoMpEntry.length);
        copyBuffer(infoMpEntry, 0, buff, value, "APP2 MPEntry-" + index);
        
        return offset;
    }

    /**
     * バッファ領域をコピーする。
     */
    private static int copyBuffer(byte[] src, int srcPos, byte[] dst, int dstPos, String tag)
    {
        return copyBuffer(src, srcPos, dst, dstPos, src.length, tag);
    }

    /**
     * バッファ領域をコピーする。
     */
    private static int copyBuffer(byte[] src, int srcPos, byte[] dst, int dstPos, int length, String tag)
    {
        /* コピー元バッファのサイズチェック */
        if(src.length < srcPos + length) {
            throw new RuntimeException("Invalid data : not enough length, " + tag);
        }
        /* コピー先バッファのサイズチェック */
        if(dst.length < dstPos + length) {
            throw new RuntimeException("Invalid heap : not enough memory, " + tag);
        }
        /* バッファコピー */
        System.arraycopy(src, srcPos, dst, dstPos, length);
        /* コピー済みサイズを返却 */
        return length;
    }

    /**
     * マーカが存在するかどうかチェックする。
     */
    private static boolean containsMarker(byte[] src, int srcPos, byte[] target, String tag) {
        /* サイズチェック */
        if(src.length < srcPos + target.length) {
            return false; /* 長さ不一致 */
        }
        /* 1byteづつチェック */
        int pos = target.length;
        while(0 < pos--) {
            if(src[srcPos + pos] != target[pos]) {
                return false; /* 不一致データあり */
            }
        }
        return true; /* 完全一致 */
    }

    /**
     * int(4byte)値をバッファにコピーする。
     */
    private void setLongValue(byte[] buff, int offset, int value)
    {
        /* little endian */
        buff[offset  ] = (byte)(value & 0xFF);
        buff[offset+1] = (byte)((value >>  8) & 0xFF);
        buff[offset+2] = (byte)((value >> 16) & 0xFF);
        buff[offset+3] = (byte)((value >> 24) & 0xFF);
    }
}
