/*
 * Copyright (C) 2011 SHARP Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.co.sharp.android.SampleCamera;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import android.hardware.Camera.Size;
import jp.co.sharp.android.hardware.CameraEx;
import jp.co.sharp.android.hardware.CameraEx.CameraInfo;
import jp.co.sharp.android.stereo3dlcd.SurfaceController;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.view.Gravity;
import android.view.MotionEvent;
import android.view.OrientationEventListener;
import android.view.Surface;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.Window;
import android.view.WindowManager;
import android.view.ViewGroup.LayoutParams;
import android.widget.LinearLayout;
import android.widget.Toast;

/**
 * カメラのプレビュー画面を表示するアクティビティである。
 * カメラの表示にはSurfaceViewを使用する。
 * メニューで指定したカメラポジションを元に、メイン/サブいずれかのカメラをオープンする。
 */
public class PreviewActivity extends Activity{

    private Preview mPreview = null;
    /* 端末方向 */
    private int mOrientation = OrientationEventListener.ORIENTATION_UNKNOWN;
    /* 端末方向通知リスナ */
    private MyOrientationEventListener mOrientationListener = null;

    /* カメラID */
    private int mCameraId;
    /* カメラデバイス */
    private CameraEx mCamera;
    /* pauseフラグ */
    private boolean isPause = false;

    /**
     *  アクティビティ作成時に呼ばれるメソッドである。
     *  Previewの作成を行う。
     */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        requestWindowFeature(Window.FEATURE_NO_TITLE);

        /* フルスクリーンにする */
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN);
        /* カメラIDをインテントより取得する */
        mCameraId = getIntent().getExtras().getInt("camera_id");

        /* 端末方向通知リスナの生成/起動 */
        mOrientationListener = new MyOrientationEventListener(this);
        mOrientationListener.enable();
    }

    @Override
    protected void onSaveInstanceState(Bundle outState) {
        outState.putInt("camera_id", mCameraId);
        super.onSaveInstanceState(outState);
    }

    @Override
    protected void onRestoreInstanceState(Bundle savedInstanceState) {
        mCameraId = savedInstanceState.getInt("camera_id");
    }

    /**
     * カメラデバイスは占有できないリソースのため、アクティビティが停止するときに
     * カメラの解放を行う。
     */
    @Override
    protected void onPause() {
        isPause = true;

        /* 端末方向通知エンジンを停止 */
        mOrientationListener.disable();

        if (mCamera != null) {
            mCamera.release();
            mCamera = null;
        }
        super.onPause();
    }

    @Override
    protected void onResume() {
        super.onResume();

       /* 端末方向通知エンジンを起動 */
        if (mOrientationListener.canDetectOrientation()){
            mOrientationListener.enable();
        }
        isPause = false;

        /* プレビューを作成 */
        LinearLayout liLayout = new LinearLayout(this);

        /* プレビューのSurfaceViewを作成 */
        mPreview = new Preview(this, mCameraId);
        /* ビューの中央にSurfaceViewを設置する */
        liLayout.setOrientation(LinearLayout.HORIZONTAL);
        liLayout.setGravity(Gravity.CENTER);
        liLayout.addView(mPreview);

        /* ビューを設定 */
        setContentView(liLayout);

    }

    @Override
    public boolean onTouchEvent(MotionEvent event) {
        switch (event.getAction()) {
        case MotionEvent.ACTION_DOWN:
            /* タッチダウンで撮影を行う */
            mPreview.takePicture();
            return true;

        default:
            break;
        }
        return super.onTouchEvent(event);
    }

    /**
     * 端末方向が変更された場合の通知を受け取るためのインタフェースである。
     */
    private class MyOrientationEventListener extends OrientationEventListener {
        public MyOrientationEventListener(Context context) {
            super(context);
        }

        @Override
        public void onOrientationChanged(int orientation) {
            if(orientation == ORIENTATION_UNKNOWN) return;

            /* 通知された端末方向を保存（90度単位で丸める） */
            mOrientation = ((orientation + 45) / 90 * 90) % 360;
        }
    }

    /**
     * カメラのプレビューを表示するクラスである。
     * SurfaceViewを継承し、SurfaceHolder.Callbackインタフェースを実装する。
     */
    class Preview extends SurfaceView implements SurfaceHolder.Callback {
        /* カメラデバイス情報 */
        private CameraInfo mCameraInfo = null;
        /* カメラパラメータ */
        private CameraEx.Parameters mParameters = null;
        /* プレビュー表示中フラグ */
        private boolean mIsPreview = false;

        /**
         * コンストラクタ
         * @param context コンテキスト
         * @param cameraId カメラID
         */
        Preview(Context context, int cameraId) {
            super(context);

            /* カメラプレビュー用にSurfaceを設定 */
            SurfaceHolder holder = getHolder();
            holder.addCallback(this);
            holder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);

            mCameraId = cameraId;
        }


        /**
         * サーフェスビュー作成時に呼ばれるメソッドである。
         * このメソッドでは、カメラAPIを用いてカメラのオープンを行う。
         * CameraExクラスのopenを呼び、引数にカメラポジションを渡すことにより、
         * オープンするカメラを切り替える。
         * オープンに成功すると、Android標準を拡張したカメラクラスが取得できるので、
         * 以降はAndroid標準のカメラクラスと同等の操作を行う。
         * @param holder サーフェスホルダ
         */
        public void surfaceCreated(SurfaceHolder holder) {

            /* pause中なので終了 */
            if(isPause) return;

            /* カメラデバイスをオープン */
            if(mCamera==null) mCamera = CameraEx.open(mCameraId); /* ★Addon用API使用箇所★ */

            /* カメラデバイス情報を取得 */
            mCameraInfo = new CameraEx.CameraInfo();
            CameraEx.getCameraInfo(mCameraId, mCameraInfo);    /* ★Addon用API使用箇所★ */
            /* 画面表示方向を取得 */
            int rotation = ((Activity)getContext()).getWindowManager()
                                .getDefaultDisplay().getRotation();
            switch (rotation) {
            case Surface.ROTATION_0: rotation = 0; break;
            case Surface.ROTATION_90: rotation = 90; break;
            case Surface.ROTATION_180: rotation = 180; break;
            case Surface.ROTATION_270: rotation = 270; break;
            default: rotation = 0; break;
            }
            /* プレビュー表示方向を設定 */
            int degrees = 0;
            switch(mCameraInfo.facing) {
            case CameraInfo.CAMERA_FACING_FRONT:
                degrees = (mCameraInfo.orientation + rotation) % 360;
                degrees = (360 - degrees) % 360; /* 180度回転 */
                break;
            case CameraInfo.CAMERA_FACING_BACK:
            default:
                degrees = (mCameraInfo.orientation - rotation + 360) % 360;
                break;
            }
            mCamera.setDisplayOrientation(degrees);
            /* Surfaceビュー表示形式設定(2D/3D) */
            switch(mCameraInfo.mode) {
            case CameraInfo.CAMERA_MODE_DOUBLE:
                new SurfaceController(this).setStereoView(true);  /* ★Addon用API使用箇所★ */
                break;
            case CameraInfo.CAMERA_MODE_SINGLE:
            default:
                new SurfaceController(this).setStereoView(false);  /* ★Addon用API使用箇所★ */
                break;
            }
            try {
                /* プレビューSurfaceを設定 */
                mCamera.setPreviewDisplay(holder);
                /* カメラデバイス種別をトースト表示 */
                showToast();
            } catch (IOException exception) {
                mCamera.release();
                mCamera = null;
            }
            System.out.println("surfaceCreated() end");
        }

        /**
         * Surfaceビュー破棄時に呼ばれるメソッドである。
         * プレビューの停止を行う。
         */
        public void surfaceDestroyed(SurfaceHolder holder) {
            /* プレビューの停止 */
            if (mCamera != null) {
                mCamera.stopPreview();
            }
        }

        /**
         * SurfaceViewに変化が起きたときに呼ばれるメソッドである。
         * カメラサイズの変更とプレビューの開始を行う。
         */
        public void surfaceChanged(SurfaceHolder holder, int format, int width, int height) {

            /* pause中なので終了 */
            if(isPause) return;

            /* カメラパラメータ取得 */
            mParameters = mCamera.getParameters();

            /* カメラに設定可能なプレビューサイズの先頭（最大）を取得する */
            Size size = mParameters.getSupportedPreviewSizes().get(0);

            /* 引数のwidth、heightが設定可能サイズより大きければ設定可能サイズに変更する */
            if (width > size.width || height > size.height) {
                width = size.width;
                height = size.height;

                //SurfaceViewのサイズを合わせる
                LayoutParams paramLayout = getLayoutParams();
                paramLayout.width = width;
                paramLayout.height = height;
                setLayoutParams(paramLayout);
            }

            /* カメラプレビューサイズ設定 */
            mParameters.setPreviewSize(width, height);

            /* カメラ撮影サイズ設定 */
            size = mParameters.getSupportedPictureSizes().get(0);
            mParameters.setPictureSize(size.width, size.height);
            /* カメラパラメータ反映 */
            mCamera.setParameters(mParameters);

            /* プレビュー開始 */
            mCamera.startPreview();
            /* プレビュー中状態に移行 */
            mIsPreview = true;
        }

        /**
         * トーストの表示を行う。
         * 表示したカメラの種類を表示する。
         */
        private void showToast() {
            try {
                /* 表示文字列取得 */
                String cameraPosition = getIntent().getExtras().getString("camera_name");
                /* トーストの表示 */
                Toast.makeText(getContext(), cameraPosition, Toast.LENGTH_LONG).show();
            } catch (IndexOutOfBoundsException e) {
                e.printStackTrace();
            }
        }

        /**
         * 撮影処理を行う。
         * カメラデバイスに対して撮影を指示し、撮影画像取得コールバックインタフェースを設定する。
         * 撮影画像を受け取ったら画像表示用のアクティビティを起動する。
         */
        public void takePicture() {
            /* プレビュー中でなければ何もせず終了 */
            if(!mIsPreview) return;

            /* 撮影画像の保存方向を設定 */
            int rotation = 0;
            if (mOrientation != OrientationEventListener.ORIENTATION_UNKNOWN) {
                switch(mCameraInfo.facing) {
                case CameraInfo.CAMERA_FACING_FRONT:
                    rotation = (mCameraInfo.orientation - mOrientation + 360) % 360;
                    break;
                case CameraInfo.CAMERA_FACING_BACK:
                default:
                    rotation = (mCameraInfo.orientation + mOrientation) % 360;
                    break;
                }
            }
            mParameters.setRotation(rotation);

            /* カメラパラメータをカメラデバイスに反映 */
            mCamera.setParameters(mParameters);

            /* 撮影画像を受け取るためのコールバックインタフェースを作成 */
            JpegPictureCallback jpeg = new JpegPictureCallback();

            /* 撮影処理をコール */
            ((CameraEx)mCamera).takePicture(null, null, null, jpeg); /* ★Addon用API使用箇所★ */

            /* takePicture()によりプレビュー状態が終了するためフラグOFF */
            mIsPreview = false;
        }

        /**
         * 撮影画像を受け取るためのコールバックインタフェース。
         * カメラデバイスから受け取った撮影画像をファイルに保存する。
         * 3D撮影画像（撮影画像が2枚）の場合はMPO形式で保存する。
         * 撮影画像はPictureManagerにて表示させる。
         */
        private class JpegPictureCallback implements CameraEx.DoublePictureCallback {
            /* 撮影画像取得コールバック関数 */
            public void onPictureTaken(byte[] left, byte[] right, CameraEx arg2) {
                if(left == null) return;

                /* 撮影画像をファイルに保存 */
                String path = createImageFile(left, right);

                /* PictureManager起動用インテントを作成 */
                Intent intent = new Intent(Intent.ACTION_VIEW);
                intent.setClassName(
                        "jp.co.sharp.android.picturemanager",
                        "jp.co.sharp.android.picturemanager.media.PictureManager");
                intent.setType(path.endsWith(".mpo") ? "image/mpo" : "image/jpeg");
                intent.setData(Uri.fromFile(new File(path)));

                /* PictureManagerを起動 */
                startActivity(intent);
            }
        }

        /**
         * バイナリデータをローカルディレクトリにファイル保存する。
         * 保存に成功した場合、ファイルパスを返却する。
         */
        private String createImageFile(byte[] left, byte[] right) {
            String name = null; /* ファイル名 */
            byte[] data = null; /* 画像データ */

            /* ファイル名と画像データを取得 */
            if(right == null) {
                /* 左画像のみの場合は、そのままJPEGファイルとして保存 */
                name = "picture" + mCameraId + "-" + System.currentTimeMillis() + ".jpg";
                data = left;
            } else {
                /* 左右とも指定された場合は、3D形式(MPO形式)に変換して保存 */
                byte[][] jpegData = {left, right};
                byte[] mpoData = new MpfUtility().createMpfData(jpegData);
                name = "picture" + mCameraId + "-"  + System.currentTimeMillis() + ".mpo";
                data = mpoData;
            }

            /* ローカルディレクトリ配下のファイルを全削除 */
            for(File file : new File(getFilesDir().getPath()).listFiles()) {
                deleteFile(file.getName());
            }

            if (Environment.MEDIA_MOUNTED.equals(Environment.getExternalStorageState())) {
                /* SDカードが挿入されていればSDカードへ保存するためのファイルパスを作成 */
                File dataFile = new File(Environment.getExternalStorageDirectory() + File.separator + name);
                try {
                    FileOutputStream fos = new FileOutputStream(dataFile);

                    fos.write(data);
                    fos.close();
                } catch (IOException e) {
                    e.printStackTrace();
                    return null;
                }
                /* ファイルパスを生成/返却 */
                return dataFile.getPath();
            } else {
                /*
                 * SDカードが未挿入の場合は、撮影後のプレビューのみ行うため
                 * 画像データをローカルディレクトリにファイル保存
                 */
                try {
                    /* 他機能からアクセスできるように MODE_WORLD_READABLE を指定 */
                    FileOutputStream fos = openFileOutput(name, Context.MODE_WORLD_READABLE);
                    fos.write(data);
                    fos.close();
                } catch (IOException e) {
                    e.printStackTrace();
                    return null;
                }
                /* ファイルパスを生成/返却 */
              return getFilesDir().getPath() + File.separator + name;
            }
        }

    }
}
